#include        <iostream>
#include        <iomanip>
#include	<stdlib.h>
#include	<string.h>
#include	<math.h>

using namespace std;

#include	"matrix.h"

/////////////////////////// protected helper functions (for constructors) //////
template <typename Type>
void Matrix<Type>::init(int row, int col, Type* array)
{
  // may be called with NULL array, but row,col must be valid (default = 4,4)

  r = row;  c = col;				// set private members r,c
  arr = new Type [r*c];				// allocate storage for arr

  // if non-NULL array, copy data from array, otherwise just set arr to zero
  if(array)	copy(array);
  else		memset(arr,0,r*c*sizeof(Type));
}

template <typename Type>
inline void Matrix<Type>::copy(Type *array)
{
  // arr must have been allocated, r,c must have been set,
  // so just copy all elements from array to arr

  if(!array)
    cerr << "copy: Warning! null array!" << endl;
  memcpy(arr,array,r*c*sizeof(Type));
}

/////////////////////////// operators: unary ///////////////////////////////////
template <typename Type>
Matrix<Type> Matrix<Type>::operator-(void)
{
	Matrix<Type>	result(r,c);

  // create 0 matrix, dim (r,c), subtract this matrix from 0 to get -ve matrix

  return(result -= this);
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator!(void)
{
	int		i,j;
	int		n = r;		// r must == c
	int		*indx = new int [n];
	Type		d, *col = new Type [n];
	Matrix<Type>	copy(*this), inverse(r,c);

  if(r == c) {
    copy.ludcmp(indx,&d);
    for(j=0; j<n; j++) {
      for(i=0; i<n; i++) col[i] = 0.0;
      col[j] = 1.0;
      copy.lubksb(indx,col);
      for(i=0; i<n; i++) inverse(i,j) = col[i];
    }
  }

  delete [] indx;  delete [] col;

  return(inverse);
}

template <typename Type>
Matrix<Type> Matrix<Type>::ludcmp(int *indx,Type *d)
{
  // Given a square matrix, this routine replaces it by the LU decomposition
  // of a rowwise permutation of itself.  indx is an output vector that
  // records the row permutation effected by the partial pivoting; d is
  // ouput as 1 or -1 depending on whether the number of row interchanges
  // was even or odd, respectively.  This routine is meant to be used in
  // combination with lubksb to solve linear equations or invert a matrix.
  //
  // From: Press, William H., Teukolsky, Saul A., Vetterling, William T., and
  //       Flannery, Brian P., ``Numerical Recipes in C:  The Art of Scientific
  //       Computing'', Cambridge University Press, (Cambridge:1992), 2nd ed.

	int	i,j,k;
	int	n = r;		// r must == c
	int	imax;
	Type	big,dum,sum,temp;
				// v stores the implicit scaling of each row
	Type	*v = new Type [n];

  *d = 1.0;			// no row interchanges yet
  for(i=0; i<n; i++) {		// loop over rows to get the impl. scaling info
    big = 0.0;
    for(j=0; j<n; j++)
      if((temp = fabsf((*this)(i,j))) > big) big = temp;
    if(big == 0.0) {
      cerr << "ludcmp: Warning! singular matrix!" << endl;
      return(*this);
    }
    // no nonzero largest element
    v[i] = 1.0/big;		// save the scaling
  }
  for(j=0; j<n; j++) {		// this is the loop over col's of Crout's method
    for(i=0; i<j; i++) {	// this is equation (2.3.12) except for i = j
      sum = (*this)(i,j);
      for(k=0; k<i; k++) sum -= (*this)(i,k) * (*this)(k,j);
      (*this)(i,j) = sum;
    }
    big = 0.0;			// init for search for largest pivot element
    for(i=j; i<n; i++) {	// this is i=j of equation (2.3.12) and
      sum = (*this)(i,j);	// i=j+1..n of equation (2.3.13)
      for(k=0; k<j; k++) sum -= (*this)(i,k) * (*this)(k,j);
      (*this)(i,j) = sum;
      if((dum = v[i]*fabsf(sum)) >= big) {
        // is the figure of merit for the pivot better than the best so far?
        big = dum;
        imax = i;
      }
    }
    if(j != imax) {		// do we need to interchange rows?
      for(k=0; k<n; k++) {	// yes, do so...
        dum = (*this)(imax,k);
        (*this)(imax,k) = (*this)(j,k);
        (*this)(j,k) = dum;
      }
      *d = -(*d);		// ...and change the parity of d
      v[imax] = v[j];		// also interchange the scale factor
    }
    indx[j] = imax;
    if((*this)(j,j) == 0.0) {
      cerr << "ludcmp: Warning! possible singular matrix!" << endl;
      (*this)(j,j) = TINY;
    }
    // if the pivot element is zero the matrix is singular (at least to the
    // precision of the algorithm).  for some applications on singular
    // matrices, it is desirable to substitute TINY for zero
    if(j != n-1) {		// now, finally, divide by the pivot element
      dum = 1.0/((*this)(j,j));
      for(i=j+1; i<n; i++) (*this)(i,j) *= dum;
    }
  }				// go back for the next column in the reduction
  delete [] v;

  return *this;
}

template <typename Type>
Matrix<Type> Matrix<Type>::lubksb(int *indx, Type *b)
{
  // Solves the set of n linear equations A . X = B.  Here (*this) square
  // matrix is assumed to be its LU decomposition, determined by the member
  // ludcmp.  indx (of dimension n, which is equal to either of r or c since
  // r == c), is input as the permutation vector returned by ludcmp. b (also
  // of dimension n) is input as the right-hand side vector B, and returns
  // with the solution vector X.  The (*this) matrix, n, and indx are not
  // modified by this member and can be left in place for successive calls
  // with different right-hand sides b.  This routine takes into account
  // the possibility that b will begin with many zero elements, so it is
  // efficient for use in matrix inversion.
  //
  // From: Press, William H., Teukolsky, Saul A., Vetterling, William T., and
  //       Flannery, Brian P., ``Numerical Recipes in C:  The Art of Scientific
  //       Computing'', Cambridge University Press, (Cambridge:1992), 2nd ed.

	int	n = r;		// r must == c
	int	i,j,ii=-1,ip;
	Type	sum;

  for(i=0; i<n; i++) {		// when ii is set to a positive value, it will
    ip = indx[i];		//   become the index of the first non-
    sum = b[ip];		//   vanishing element of b.  we now do the
    b[ip] = b[i];		//   forward substitution, equation (2.3.6).
          			//   the only new wrinkle is to unsramble the
				//   permutation as we go
    if(ii >= 0)
      for(j=ii; j<=i-1; j++) sum -= (*this)(i,j) * b[j];
    else if(sum)		// a nonzero element was encountered, so from
      ii = i;			//   now on we will have to do the sums in
    b[i] = sum;			//   the loop above
  }
  for(i=n-1; i>=0; i--) {	// now do the backsubstitution, eq. (2.3.7)
    sum = b[i];
    for(j=i+1; j<n; j++) sum -= (*this)(i,j) * b[j];
    b[i] =                      // store a component of the solution vector X
      sum/(*this)(i,i);
  }				// all done!

  return *this;
}

/////////////////////////// operators: (const Matrix&, const Matrix&) //////////
template <typename Type>
int Matrix<Type>::closeto(const Matrix<Type>& rhs)
{
	float	precision=10.01;

  if(this == &rhs)				// compare _pointers_ !
    return(1);
  else {
    if((r != rhs.r) || (c != rhs.c))
      return(0);
    else {
      for(int i=0; i<r; i++)
        for(int j=0; j<c; j++)
          if(fabs((*this)(i,j) - rhs(i,j)) > precision)
            return(0);
    }
  }
  return(1);					// if we're here, rhs is "close"
}

template <typename Type>
int Matrix<Type>::operator<(const Matrix<Type>& rhs)
{
  if(this == &rhs)				// compare _pointers_ !
    return(0);
  else {
    if((r != rhs.r) || (c != rhs.c))
      return(0);
    else {
      for(int i=0; i<r; i++)
        for(int j=0; j<c; j++)
          if((*this)(i,j) >= rhs(i,j))
            return(0);
    }
  }
  return(1);					// if we get here, lhs < rhs
}

template <typename Type>
int Matrix<Type>::operator>(const Matrix<Type>& rhs)
{
  if(this == &rhs)				// compare _pointers_ !
    return(0);
  else {
    if((r != rhs.r) || (c != rhs.c))
      return(0);
    else {
      for(int i=0; i<r; i++)
        for(int j=0; j<c; j++)
          if((*this)(i,j) <= rhs(i,j))
            return(0);
    }
  }
  return(1);					// if we get here, lhs > rhs
}

template <typename Type>
int Matrix<Type>::operator==(const Matrix<Type>& rhs)
{
	float	precision=0.01;

  if(this == &rhs)				// compare _pointers_ !
    return(1);
  else {
    if((r != rhs.r) || (c != rhs.c))
      return(0);
    else {
      for(int i=0; i<r; i++)
        for(int j=0; j<c; j++)
          if(fabs((*this)(i,j) - rhs(i,j)) > precision)
            return(0);
    }
  }
  return(1);					// if we get here, lhs == rhs
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator=(const Matrix<Type>& rhs)
{
  if(this != &rhs) {				// compare _pointers_ !
    if((r != rhs.r) || (c != rhs.c)) {
      delete [] arr;				// arr already exists, nuke it!
      r = rhs.r;  c = rhs.c;
      arr = new Type [r*c];
    }
    copy(rhs.arr);
  }
  return *this;
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator*=(const Matrix<Type>& rhs)
{
  return((*this) = (*this) * rhs);
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator*(const Matrix<Type>& rhs)
{
	Matrix<Type>	result(r,rhs.c);

  if(c == rhs.r) {
    for(int i=0; i<result.r; i++)
      for(int j=0; j<result.c; j++)
        for(int k=0; k<c; k++)
          result(i,j) += (*this)(i,k) * rhs(k,j);
  }
  return result;
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator/=(Matrix<Type>& rhs)
{
  return((*this) = (*this) * !rhs);
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator/(Matrix<Type>& rhs)
{
	Matrix<Type>	result(r,rhs.c);

  return(result = (*this) * !rhs);
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator+=(const Matrix<Type>& rhs)
{
	int	rows,cols;

  rows = r <= rhs.r ? r : rhs.r;
  cols = c <= rhs.c ? c : rhs.c;

  for(int i=0; i<rows; i++)
    for(int j=0; j<cols; j++)
      (*this)(i,j) += rhs(i,j);

  return *this;
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator+(const Matrix<Type>& rhs)
{
	Matrix<Type>	result(this);

  return(result += rhs);
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator-=(const Matrix<Type>& rhs)
{
	int	rows,cols;

  rows = r <= rhs.r ? r : rhs.r;
  cols = c <= rhs.c ? c : rhs.c;

  for(int i=0; i<rows; i++)
    for(int j=0; j<cols; j++)
      (*this)(i,j) -= rhs(i,j);

  return *this;
}


template <typename Type>
Matrix<Type> Matrix<Type>::operator-(const Matrix<Type>& rhs)
{
	Matrix<Type>	result(this);

  return(result -= rhs);
}


/////////////////////////// operators: (const Matrix&, scalar) /////////////////
template <typename Type>
Matrix<Type> Matrix<Type>::operator*=(const Type& rhs)
{
  for(int i=0; i<r; i++)
    for(int j=0; j<c; j++)
      (*this)(i,j) *= rhs;

  return *this;
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator*(const Type& rhs)
{
	Matrix<Type>	result(this);

  return(result *= rhs);
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator/=(const Type& rhs)
{
  for(int i=0; i<r; i++)
    for(int j=0; j<c; j++)
      (*this)(i,j) /= rhs;

  return *this;
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator/(const Type& rhs)
{
	Matrix<Type>	result(this);

  return(result /= rhs);
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator+=(const Type& rhs)
{
  for(int i=0; i<r; i++)
    for(int j=0; j<c; j++)
      (*this)(i,j) += rhs;

  return *this;
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator+(const Type& rhs)
{
	Matrix<Type>	result(this);

  return(result += rhs);
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator-=(const Type& rhs)
{
  for(int i=0; i<r; i++)
    for(int j=0; j<c; j++)
      (*this)(i,j) -= rhs;

  return *this;
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator-(const Type& rhs)
{
	Matrix<Type>	result(this);

  return(result -= rhs);
}


/////////////////////////// friend operators: (const scalar&, const Matrix&) ///
template <typename Type>
Matrix<Type> operator*(const Type& lhs,const Matrix<Type>& rhs)
{
	Matrix<Type>	result(rhs);

  return(result *= lhs);
}

template <typename Type>
Matrix<Type> operator/(const Type& lhs,const Matrix<Type>& rhs)
{
	double		prec = 0.001;
	Matrix<Type>	result(rhs);

  // tensor product?  this routine calculates (scalar / matrix), giving
  // [ scalar/element scalar/element ... ]
  // [ scalar/element scalar/element ... ]
  // [ ...                           ... ]

  for(int i=0; i<result.rows(); i++)
    for(int j=0; j<result.cols(); j++)
      result(i,j) =
        // check for div by zero
        fabs(result(i,j) - prec) <= prec ? HUGE : lhs/result(i,j);

  return(result);
}

template <typename Type>
Matrix<Type> operator+(const Type& lhs,const Matrix<Type>& rhs)
{
	Matrix<Type>	result(rhs);

  return(result += lhs);
}

template <typename Type>
Matrix<Type> operator-(const Type& lhs,const Matrix<Type>& rhs)
{
	Matrix<Type>	result(rhs);

  result = -result;
  return(result += lhs);
}


/////////////////////////// friends ////////////////////////////////////////////
template <typename Type>
ostream& operator<<(ostream& s,const Matrix<Type>& rhs)
{
  s.setf(ios::fixed,ios::floatfield);
  s.precision(2);
  for(int i=0; i<rhs.r; i++) {
    s << "[";
    for(int j=0; j<rhs.c; j++) {
      s.width(6);
      j == rhs.c-1 ? s << rhs(i,j) : s << rhs(i,j) << ", ";
    }
    s << "]" << endl;
  }
  
  return s;
}

/////////////////////////// public member functions ////////////////////////////
template <typename Type>
void Matrix<Type>::one(void)
{
  for(int i=0; i<r; i++)
    for(int j=0; j<c; j++)
      (*this)(i,j) = static_cast<Type>(1.0);
}

template <typename Type>
void Matrix<Type>::zero(void)
{
  for(int i=0; i<r; i++)
    for(int j=0; j<c; j++)
      (*this)(i,j) = static_cast<Type>(0.0);
}

template <typename Type>
Matrix<Type> Matrix<Type>::clamp(Type lo, Type hi)
{
  for(int i=0; i<r; i++) {
    for(int j=0; j<c; j++) {
      if( (*this)(i,j) < lo ) (*this)(i,j) = lo;
      if( (*this)(i,j) > hi ) (*this)(i,j) = hi;
    }
  }
  return *this;
}

template <typename Type>
void Matrix<Type>::identity(void)
{
  for(int i=0; i<r; i++)
    for(int j=0; j<c; j++)
      i == j ? (*this)(i,j) = static_cast<Type>(1.0) :
               (*this)(i,j) = static_cast<Type>(0.0);
}

template <typename Type>
Type Matrix<Type>::dot(const Matrix<Type>& rhs)
{
        Type	sum=0.0;
	int	rows,cols;

  rows = r <= rhs.r ? r : rhs.r;
  cols = c <= rhs.c ? c : rhs.c;

  for(int i=0; i<rows; i++)
    for(int j=0; j<cols; j++)
      sum += (*this)(i,j) * rhs(i,j);
  return sum;
}

template <typename Type>
Type Matrix<Type>::trace(void) const
{
        Type   sum=0.0;

  for(int i=0; i<r; i++)
    for(int j=0; j<c; j++)
      if(i==j) sum += (*this)(i,j);

  return sum;
}

template <typename Type>
Matrix<Type> Matrix<Type>::transpose(void)
{
	Matrix<Type>	result(r,c);

  for(int i=0;i<r;i++)
    for(int j=0;j<c;j++)
      result(i,j) = (*this)(j,i);

  return(result);
}

