#ifndef MATRIX_H
#define MATRIX_H

/* Wraparound defines:  note that i,j can be (i+ni), (i-ni), (j+nj), (j-nj)
   which allows the indices i,j to go either left or right in the image
   row */
#define	I(i,r)			(((r)+(i))%(r))
#define	J(j,c)			(((c)+(j))%(c))
#define	M2A(i,j,r,c)		(I((i),(r))*(c) + J((j),(c)))

#define TINY			1.0e-20;

template <typename Type>
class Matrix {
	public:
	// constructors (overloaded)
	Matrix(int row=4,int col=4)		{init(row,col,NULL);}
	Matrix(Type* array, int row, int col)	{init(row,col,array);}
	Matrix(const Matrix& rhs)		{init(rhs.r,rhs.c,rhs.arr);}
	Matrix(Matrix* rhs)			{init(rhs->r,rhs->c,rhs->arr);}

	// destructors
	virtual ~Matrix()				{delete [] arr;}

	// operators: dereferencing (Fortran-like indeces), with wraparound
	Type& operator()(int i,int j)
	  { return(arr[((r+i)%r)*c + ((c+j)%c)]); }
	Type& operator()(int i,int j) const
	  { return(arr[((r+i)%r)*c + ((c+j)%c)]); }

	// operators: unary
	Matrix operator-(void);
	Matrix operator!(void);			// matrix inverse

	// operators: (const Matrix&, const Matrix&)
	int operator==(const Matrix& rhs);
	int operator<(const Matrix& rhs);
	int operator>(const Matrix& rhs);
	Matrix operator=(const Matrix& rhs);
	Matrix operator*=(const Matrix& rhs);
	Matrix operator*(const Matrix& rhs);
	Matrix operator/=(Matrix& rhs);
	Matrix operator/(Matrix& rhs);
	Matrix operator+=(const Matrix& rhs);
	Matrix operator+(const Matrix& rhs);
	Matrix operator-=(const Matrix& rhs);
	Matrix operator-(const Matrix& rhs);

	// operators: (const Matrix&, Matrix*)
	int operator==(Matrix* rhs)		{ return((*this) == (*rhs)); }
	int operator<(Matrix* rhs)		{ return((*this) == (*rhs)); }
	int operator>(Matrix* rhs)		{ return((*this) == (*rhs)); }
	Matrix operator=(Matrix* rhs)		{ return((*this)  = (*rhs)); }
	Matrix operator*=(Matrix* rhs)		{ return((*this) *= (*rhs)); }
	Matrix operator*(Matrix* rhs)		{ return((*this) *  (*rhs)); }
	Matrix operator/=(Matrix* rhs)		{ return((*this) /= (*rhs)); }
	Matrix operator/(Matrix* rhs)		{ return((*this) /  (*rhs)); }
	Matrix operator+=(Matrix* rhs)		{ return((*this) += (*rhs)); }
	Matrix operator+(Matrix* rhs)		{ return((*this) +  (*rhs)); }
	Matrix operator-=(Matrix* rhs)		{ return((*this) -= (*rhs)); }
	Matrix operator-(Matrix* rhs)		{ return((*this) -  (*rhs)); }

	// operators: (const Matrix&, scalar)
	Matrix operator*=(const Type& rhs);
	Matrix operator*(const Type& rhs);
	Matrix operator/=(const Type& rhs);
	Matrix operator/(const Type& rhs);
	Matrix operator+=(const Type& rhs);
	Matrix operator+(const Type& rhs);
	Matrix operator-=(const Type& rhs);
	Matrix operator-(const Type& rhs);

	// friend operators: (const scalar&, const Matrix&)
	friend Matrix operator*(const Type& lhs,const Matrix &rhs);
	friend Matrix operator/(const Type& lhs,const Matrix &rhs);
	friend Matrix operator+(const Type& lhs,const Matrix &rhs);
	friend Matrix operator-(const Type& lhs,const Matrix &rhs);

	// friends
	friend ostream& operator<<(ostream& s, const Matrix& rhs);
	friend ostream& operator<<(ostream& s, Matrix *rhs)
	  { return(s << (*rhs)); }

	// public member functions
	void	one(void);
	void	zero(void);
	void	identity(void);
	Matrix	clamp(Type lo, Type hi);
	Matrix	transpose(void);
	Type*	v()				{ return(arr); }
	Type	dot(const Matrix& rhs);
	Type	dot(const Matrix* rhs)		{ return(dot((*rhs))); }
	int	closeto(const Matrix& rhs);
	int	closeto(const Matrix* rhs)	{ return(closeto((*rhs))); }
	Type	length(void)			{ return(sqrt(dot(this))); }
	Matrix	norm(void)			{ return(*this / length()); }
	int	rows(void) const		{ return(r); }
	int	cols(void) const		{ return(c); }
	Type	trace(void) const;

	// protected: available to this (base) class and subs. derived classes
	protected:
	void	init(int row,int col,Type* array);
inline	void	copy(Type* array);
	int	r,c;
	Type	*arr;

	// private: only available to this class
	private:
	Matrix	ludcmp(int *indx,Type *d);
	Matrix	lubksb(int *indx,Type *b);
};

#endif
