/****************************************************************************
** $Id: qt/gltexobj.cpp   3.1.2   edited Nov 8 2002 $
**
** Copyright (C) 1992-2002 Trolltech AS.  All rights reserved.
**
** This file is part of an example program for Qt.  This example
** program may be used, distributed and modified without limitation.
**
*****************************************************************************/

/****************************************************************************
**
** This is a simple QGLWidget demonstrating the use of QImages for textures.
**
** Much of the GL code is inspired by the 'spectex' and 'texcyl' 
** public domain demo programs by Brian Paul.
**
****************************************************************************/

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>

#ifdef ANM_OSX
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#include <OpenGL/glext.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>
#endif
#include <Cg/cg.h>
#include <Cg/cgGL.h>

#include <qimage.h>
#include <qtimer.h>
#include <qevent.h>
#include <qfiledialog.h>
#include <qstring.h>

#include "gltexobj.h"

#include "utils.h"

const int textures=2;
ulong texID[] = {GL_TEXTURE0_ARB,GL_TEXTURE1_ARB};

extern "C" {

static void handleCgError();

}

//const int redrawWait = 50;
const int redrawWait = 0;

/*!
  Create a GLTexobj widget
*/

GLTexobj::GLTexobj( QWidget* parent, const char* name ) :
	QGLWidget( parent, name ), \
	dx(0.0), \
	dy(0.0), \
	fps(0.0), \
	fps_samples(0.0), \
	fps_avg(0.0), \
	lod(4)
{
	uchar	red,green,blue,alpha=0xff;
	float	g,sigma=100.0;
	ofstream ofs;

// FOR mouse CONTROL, comment these 3 lines out
//timer = new QTimer( this );
//connect( timer, SIGNAL(timeout()), SLOT(update()) );
//timer->start( redrawWait, FALSE );

  img = new Ppm(512,512,0);
  ofs.open("image.ppm");
  ofs << img;
  ofs.close();

  mask = new Ppm(img->rows(),img->cols(),0);
  for(int i=0; i<mask->rows(); ++i) {
    for(int j=0; j<mask->cols(); ++j) {
      g = gauss(i-mask->rows()/2,j-mask->cols()/2,sigma);
      red = green = blue = alpha = static_cast<uchar>(g * 255.0);
#ifdef ANM_OSX
      // store in RGBA format
        (*mask)(i,j) = ((red<<24) | (green<<16) | (blue<<8) | alpha);
#else
      // store in ABGR format (old way; for MESA)
        (*mask)(i,j) = ((alpha<<24) | (blue<<16) | (green<<8) | red);
#endif
    }
  }
  ofs.open("mask.ppm");
  ofs << mask;
  ofs.close();
}


/*!
  Release allocated resources
*/

GLTexobj::~GLTexobj()
{
  makeCurrent();
}


/*!
  Set up the OpenGL rendering state, and define display list
*/

void GLTexobj::initializeGL()
{
	const GLubyte*	v = glGetString(GL_VENDOR);

  std::cerr << "GL: " << glGetString(GL_VERSION) << std::endl;

  if( (v[0] == 'N') && (v[1] == 'V') )		vendor = NVIDIA;
  else if( (v[0] == 'A') && (v[1] == 'T') )	vendor = ATI;
  else						vendor = NVIDIA;

  cgSetErrorCallback(handleCgError);
  context = cgCreateContext();

  if(!haveLoadedPrograms) {
    chooseCgProfiles();
    loadCgPrograms();
    haveLoadedPrograms = true;
  }

  glGenTextures(textures,texNames);

  imageTexture();
  imageMask();

  // specify clear color; (0.3,0.4,0.6,1.0) is a nice "slate blue"
  glClearColor( 0.0, 0.0, 0.0, 0.0 );
}


extern "C" {

static void handleCgError()
{
  cerr << "Cg error: " << cgGetErrorString(cgGetError()) << endl;
  exit(1);
}

}

/*!
  Set up the OpenGL view port, matrix mode, etc.
*/

void GLTexobj::resizeGL( int w, int h )
{
  glViewport(0,0,w,h);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0.0,w,0.0,h);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}


/*!
  Paint the texobj. The actual openGL commands for drawing the texobj are
  performed here.
*/

void GLTexobj::paintGL()
{
	float	lx,ly;
	float	lw,lh;

  glDrawBuffer(GL_BACK);
  glClear(GL_COLOR_BUFFER_BIT);

  // Cg
  // now make sure that the vertex and fragment programs, loaded
  // in LoadCgPrograms() are bound.
  cgGLEnableProfile(vertexProfile);
  cgGLEnableProfile(fragmentProfile);
  cgGLBindProgram(vertexProgram);
  cgGLBindProgram(fragmentProgram);
  
  // Bind uniform parameters to vertex shader
  cgGLSetStateMatrixParameter(cgGetNamedParameter(vertexProgram,
                              "ModelViewProj"),
                              CG_GL_MODELVIEW_PROJECTION_MATRIX,
                              CG_GL_MATRIX_IDENTITY);
  cgGLSetStateMatrixParameter(cgGetNamedParameter(vertexProgram,
                              "ModelView"),
                              CG_GL_MODELVIEW_MATRIX,
                              CG_GL_MATRIX_IDENTITY);
  cgGLSetStateMatrixParameter(cgGetNamedParameter(vertexProgram,
                              "ModelViewIT"),
                              CG_GL_MODELVIEW_MATRIX,
                              CG_GL_MATRIX_INVERSE_TRANSPOSE);

  cgGLSetParameter1f(cgGetNamedParameter(fragmentProgram,"min_lod"),(float)lod);

  // texID[0]: image
  cgGLEnableTextureParameter(cgGetNamedParameter(fragmentProgram,"img_tex"));

  // texID[1]: mask
  cgGLEnableTextureParameter(cgGetNamedParameter(fragmentProgram,"deg_tex"));

  lx = (float)dx;
  ly = (float)dy;
//cerr << "(lx,ly): (" << lx << "," << ly << ")" << endl;
  // not sure how Nathan's 'aspect ratio' calculation works here yet...
  // (1,1) gives the proper offset for square windows and square masks
  // matching the window size...
//lw = (float)width() / mask->cols();
//lh = (float)height() / mask->rows();
  lw = 1.0;
  lh = 1.0;
//cerr << "(lw,lh): (" << lw << "," << lh << ")" << endl;

  // render
  glBegin(GL_QUADS);
  {
    glMultiTexCoord2fARB(texID[0],0,0);
    glMultiTexCoord2fARB(texID[1],lx,ly);
    glVertex2f(0,0);

    glMultiTexCoord2fARB(texID[0],1,0);
    glMultiTexCoord2fARB(texID[1],lx+lw,ly);
    glVertex2f(width(),0);

    glMultiTexCoord2fARB(texID[0],1,1);
    glMultiTexCoord2fARB(texID[1],lx+lw,ly+lh);
    glVertex2f(width(),height());

    glMultiTexCoord2fARB(texID[0],0,1);
    glMultiTexCoord2fARB(texID[1],lx,ly+lh);
    glVertex2f(0,height());
  }
  glEnd();

  cgGLDisableTextureParameter(cgGetNamedParameter(fragmentProgram,"deg_tex"));

  cgGLDisableTextureParameter(cgGetNamedParameter(fragmentProgram,"img_tex"));

  cgGLDisableProfile(fragmentProfile);
  cgGLDisableProfile(vertexProfile);

  swapBuffers();
}


void GLTexobj::mouseMoveEvent(QMouseEvent* e)
{
//std::cerr << "(" << e->x() << "," << e->y() << ")" << std::endl;
  dx = width()/2 - e->x();
  dy = e->y() - height()/2;
  dx /= width();
  dy /= height();

  e->accept();
  updateGL();
}

void GLTexobj::update()
{
//	static Timer	timer;

//std::cerr << "updating..." << std::endl;

// can do timing here...

}

void GLTexobj::imageLOD(int ilod)
{
  lod = ilod;

  updateGL();
}

void GLTexobj::imageDeg(int smoothing)
{
  // GL_TEXTURE_MIN_FILTER options:
  //                                                       (Manhattan dist.)
  //                                   intra-map interp    inter-map interp
  //   GL_NEAREST_MIPMAP_NEAREST:             no                  no
  //   GL_LINEAR_MIPMAP_NEAREST              yes                  no
  //   GL_NEAREST_MIPMAP_LINEAR               no                 yes
  //   GL_LINEAR_MIPMAP_LINEAR               yes                 yes
  // GL_TEXTURE_MAG_FILTER options:
  //   GL_NEAREST                             no                 n/a 
  //   GL_LINEAR                             yes                 n/a
  glActiveTextureARB(texID[0]);
  switch(smoothing) {
    case 1:
      // no smoothing
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                    GL_NEAREST_MIPMAP_NEAREST);
      break;
    case 2:
      // intra-map
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                    GL_LINEAR_MIPMAP_NEAREST);
      break;
    case 3:
      // inter-map
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                    GL_NEAREST_MIPMAP_LINEAR);
      break;
    case 4:
    default:
      // both
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                    GL_LINEAR_MIPMAP_LINEAR);
      break;
  }
  updateGL();
}

void GLTexobj::imageOpen()
{
    QString	qfilename = QFileDialog::getOpenFileName( QString::null,
                                                          QString::null,
                                                          this );
  if(!qfilename.isEmpty()) {
    std::ifstream ifs(qfilename.ascii());
    if(!ifs)
      cerr << "Warning: can't open " << qfilename << endl;
    else {
      ifs >> img;
      ifs.close();

      imageTexture();

//    I don't think we need to do this...
//    mask->resize(img->rows(),img->cols());
//    imageMask();
    }
  }
}


void GLTexobj::openColorDeg()
{
    QString     qfilename = QFileDialog::getOpenFileName( QString::null,
                                                          QString::null,
                                                          this );
    Ppm                 *maskppm;
    register uchar      blue,green,red,alpha;
  
  if(!qfilename.isEmpty()) {
    std::ifstream ifs(qfilename.ascii());
    if(!ifs)
      cerr << "Warning: can't open " << qfilename << endl;
    else {
  
      maskppm = new Ppm(512,512,(ushort)0);
  
      ifs >> maskppm;
      ifs.close();
  
      maskppm->resize(mask->rows(),mask->cols());
  
      for(int i=0; i<mask->rows(); ++i) {
        for(int j=0; j<mask->cols(); ++j) {
  
          red       = (uchar)LONG_R((*maskppm)(i,j));
          green     = (uchar)LONG_G((*maskppm)(i,j));
          blue      = (uchar)LONG_B((*maskppm)(i,j));
          alpha     = (uchar)LONG_A((*mask)(i,j));
          
#ifdef ANM_OSX
          // store in RGBA format
          (*mask)(i,j) = ((red<<24) | (green<<16) | (blue<<8) | alpha);
#else
          // store in ABGR format (old way; for MESA)
          (*mask)(i,j) = ((alpha<<24) | (blue<<16) | (green<<8) | red);
#endif
        }  
      }

//    imageTexture();
      imageMask();

      delete maskppm;

    }
  }
} 


void GLTexobj::openSpatDeg()
{
    QString     qfilename = QFileDialog::getOpenFileName( QString::null,
                                                          QString::null,
                                                          this );
    Pgm                 *maskpgm;
    register uchar      blue,green,red,alpha;
    
  if(!qfilename.isEmpty()) {
    std::ifstream ifs(qfilename.ascii());
    if(!ifs)
      cerr << "Warning: can't open " << qfilename << endl;
    else {

      maskpgm = new Pgm(512,512,(ushort)0);

      ifs >> maskpgm;
      ifs.close();

      maskpgm->resize(mask->rows(),mask->cols());
  
      for(int i=0; i<mask->rows(); ++i) {
        for(int j=0; j<mask->cols(); ++j) {

          red       = (uchar)LONG_R((*mask)(i,j));
          green     = (uchar)LONG_G((*mask)(i,j));
          blue      = (uchar)LONG_B((*mask)(i,j));
          alpha     = (uchar)SHORT_L((*maskpgm)(i,j));

#ifdef ANM_OSX
          // store in RGBA format
          (*mask)(i,j) = ((red<<24) | (green<<16) | (blue<<8) | alpha);
#else
          // store in ABGR format (old way; for MESA)
          (*mask)(i,j) = ((alpha<<24) | (blue<<16) | (green<<8) | red);
#endif
        } 
      }

//    imageTexture();
      imageMask();

      delete maskpgm;

    }
  }
}



void GLTexobj::imageTexture()
{
	GLint	error;

  if(glIsTexture(texNames[0])) {
    glDeleteTextures(1,&texNames[0]);
    glGenTextures(1,&texNames[0]);
  }

  // texture 0 is the 'image'
  glBindTexture(GL_TEXTURE_2D,texNames[0]);
  // set up texture env
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  error = gluBuild2DMipmaps(GL_TEXTURE_2D,
               GL_RGBA,                     // internalFormat
               img->cols(),img->rows(),     // width,height
               GL_RGBA,                     // format
               GL_UNSIGNED_BYTE,            // type
               img->data());                // texels
  if(error)	std::cerr << gluErrorString(error) << std::endl;
  else		std::cerr << "gluBuild2DMipmaps succeeded." << std::endl;
  // tell Cg which texture handle should be associated with the sampler2D
  cgGLSetTextureParameter(cgGetNamedParameter(fragmentProgram, "img_tex"),
                          texNames[0]);
}

void GLTexobj::imageMask()
{
  if(glIsTexture(texNames[1])) {
    glDeleteTextures(1,&texNames[1]);
    glGenTextures(1,&texNames[1]);
  }

  // texture 1 is the mask
  glBindTexture(GL_TEXTURE_2D,texNames[1]);
  // set up texture env
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  glTexImage2D(GL_TEXTURE_2D,     // target
               0,                           // level
               GL_RGBA,                     // internalFormat
               mask->cols(),mask->rows(),   // width,height
               0,                           // border
               GL_RGBA,                     // format
               GL_UNSIGNED_BYTE,            // type
               mask->data());               // texels
  // tell Cg which texture handle should be associated with the sampler2D
  cgGLSetTextureParameter(cgGetNamedParameter(fragmentProgram, "deg_tex"),
                          texNames[1]);
}

void GLTexobj::chooseCgProfiles()
{

//vertexProfile = cgGLGetLatestProfile(CG_GL_VERTEX);
//fragmentProfile = cgGLGetLatestProfile(CG_GL_FRAGMENT);
//cgGLSetOptimalOptions(vertexProfile);
//cgGLSetOptimalOptions(fragmentProfile);
//
//return;

  // Make sure that the appropriate profiles are available on the
  // user's system.
  if(cgGLIsProfileSupported(CG_PROFILE_ARBVP1)) {
    vertexProfile = CG_PROFILE_ARBVP1;
    cerr << "Cg vertex profile: arbvp1" << endl;
  } else if(cgGLIsProfileSupported(CG_PROFILE_VP30)) {
    // try VP30
    vertexProfile = CG_PROFILE_VP30;
    cerr << "Cg vertex profile: vp30" << endl;
  } else if (cgGLIsProfileSupported(CG_PROFILE_VP20)) {
    // try VP20
    vertexProfile = CG_PROFILE_VP20;
    cerr << "Cg vertex profile: vp20" << endl;
  } else {
    cerr << "Neither arbvp1, vp30, or vp20 vertex profiles supported";
    cerr << "on this system" << endl;
    exit(1);
  }

  if(cgGLIsProfileSupported(CG_PROFILE_ARBFP1)) {
    fragmentProfile = CG_PROFILE_ARBFP1;
    cerr << "Cg fragment profile: arbfp1" << endl;
  } else if (cgGLIsProfileSupported(CG_PROFILE_FP30)) {
    // try FP30
    fragmentProfile = CG_PROFILE_FP30;
    cerr << "Cg fragment profile: fp30" << endl;
  } else if (cgGLIsProfileSupported(CG_PROFILE_FP20)) {
    // try FP20
    fragmentProfile = CG_PROFILE_FP20;
    cerr << "Cg fragment profile: fp20" << endl;
  } else {
    cerr << "Neither arbfp1, fp30, or fp20 fragment profiles supported";
    cerr << "on this system" << endl;
    exit(1);
  }
}

void GLTexobj::loadCgPrograms()
{
  assert(cgIsContext(context));

  // load and compile the vertex program from "demo_vert.cg"
  // hold on to the handle to it that is returned.
  cerr << "Cg vertex program creating on this system" << endl;
  vertexProgram = cgCreateProgramFromFile(context,
                                          CG_SOURCE,
                                          "demo_vert.cg",
                                          vertexProfile, NULL, NULL);
  if(!cgIsProgramCompiled(vertexProgram)) {
    cerr << "Cg vertex program compiling on this system" << endl;
    cgCompileProgram(vertexProgram);
    cerr << "Cg vertex program compiled on this system" << endl;
  }

  // enable the appropriate vertex profile and load the vertex program.
  cgGLEnableProfile(vertexProfile);
  cerr << "Cg vertex program loading on this system" << endl;
  cgGLLoadProgram(vertexProgram);

  // and similarly set things up for the fragment program.
  cerr << "Cg fragment program creating on this system" << endl;
  fragmentProgram = cgCreateProgramFromFile(context,
                                            CG_SOURCE,
                                            "demo_frag.cg",
                                            fragmentProfile, NULL, NULL);
  if(!cgIsProgramCompiled(fragmentProgram)) {
    cerr << "Cg fragment program compiling on this system" << endl;
    cgCompileProgram(fragmentProgram);
    cerr << "Cg fragment program compiled on this system" << endl;
  }

  cgGLEnableProfile(fragmentProfile);
  cerr << "Cg fragment program loading on this system" << endl;
  cgGLLoadProgram(fragmentProgram);
}
